/** Something that we can write to. */
interface Writable {
	write(chunk: any): any
	end?(cb?: () => void): void
	close?(): Promise<void>
}

export interface TransformConfig {
	/** Where is this Transform piping to? */
	pipes: Array<Writable>
}

export interface TransformOptions extends Partial<TransformConfig> {
	[key: string]: any
}

/**
 * Transform.
 * This is a helper for our transforms to be able to process the written log data more easily.
 * All the need to do is extend this class and add their own `format` method.
 * @param args forwarded to {@link Logger#setConfig}
 * @example
 * ``` javascript
 * import {inspect} from 'util'
 * import {Logger, Transform} from 'caterpillar'
 * class Pretty extends Transform {
 * 	format (entry) {
 * 		return inspect(entry, {colors: true})
 * 	}
 * }
 * new Logger()
 * 	.pipe(new Pretty())
 * 	.pipe(process.stdout)
 * 	.log('note', 'cool times', 5)
 * ```
 */
export class Transform implements Writable {
	/** You should specify this with the interface for your own configuration which extends the `TransformConfig`. */
	config: TransformConfig

	/**
	 * Dereference the configuration and apply the pipes.
	 * You should extend this, calling super with the opts, and then applying your own options to `this.config`.
	 */
	constructor(opts?: TransformOptions) {
		this.config = {
			pipes: opts?.pipes ?? [],
		}
	}

	/**
	 * Format the current entry representation.
	 * Your transformer should extend this.
	 */
	format(message: any): any {
		return message
	}

	/** Pipe future log entries into a caterpillar transform or a stream. */
	pipe<T extends Writable>(to: T) {
		this.config.pipes.push(to)
		return to
	}

	/** Write to the child pipes. */
	write(chunk: any) {
		const data = this.format(chunk)
		// exclude filtered entries
		if (data === null) return
		// pipe to child transforms and streams
		for (const pipe of this.config.pipes) {
			if (pipe instanceof Transform || typeof data === 'string') {
				// compatibility with caterpillar transforms
				pipe.write(data)
			} else {
				// compatibility with node streams
				pipe.write(JSON.stringify(data))
			}
		}
	}

	/** Close the child pipes. */
	async close(): Promise<void> {
		await Promise.all(
			this.config.pipes.map((pipe) => {
				if (pipe.close) {
					return pipe.close()
				} else if (pipe.end) {
					return new Promise(function (resolve) {
						pipe.end!(resolve)
					})
				} else {
					return Promise.resolve()
				}
			})
		)
	}

	/* Callback alias for close */
	end(cb?: () => void) {
		this.close().finally(cb)
	}
}
