/**
 * Transform.
 * This is a helper for our transforms to be able to process the written log data more easily.
 * All the need to do is extend this class and add their own `format` method.
 * @param args forwarded to {@link Logger#setConfig}
 * @example
 * ``` javascript
 * import {inspect} from 'util'
 * import {Logger, Transform} from 'caterpillar'
 * class Pretty extends Transform {
 * 	format (entry) {
 * 		return inspect(entry, {colors: true})
 * 	}
 * }
 * new Logger()
 * 	.pipe(new Pretty())
 * 	.pipe(process.stdout)
 * 	.log('note', 'cool times', 5)
 * ```
 */
export class Transform {
	/**
	 * Dereference the configuration and apply the pipes.
	 * You should extend this, calling super with the opts, and then applying your own options to `this.config`.
	 */
	constructor(opts) {
		this.config = {
			pipes: opts?.pipes ?? [],
		}
	}
	/**
	 * Format the current entry representation.
	 * Your transformer should extend this.
	 */
	format(message) {
		return message
	}
	/** Pipe future log entries into a caterpillar transform or a stream. */
	pipe(to) {
		this.config.pipes.push(to)
		return to
	}
	/** Write to the child pipes. */
	write(chunk) {
		const data = this.format(chunk)
		// exclude filtered entries
		if (data === null) return
		// pipe to child transforms and streams
		for (const pipe of this.config.pipes) {
			if (pipe instanceof Transform || typeof data === 'string') {
				// compatibility with caterpillar transforms
				pipe.write(data)
			} else {
				// compatibility with node streams
				pipe.write(JSON.stringify(data))
			}
		}
	}
	/** Close the child pipes. */
	async close() {
		await Promise.all(
			this.config.pipes.map((pipe) => {
				if (pipe.close) {
					return pipe.close()
				} else if (pipe.end) {
					return new Promise(function (resolve) {
						pipe.end(resolve)
					})
				} else {
					return Promise.resolve()
				}
			})
		)
	}
	/* Callback alias for close */
	end(cb) {
		this.close().finally(cb)
	}
}
